@@ -8,6 +8,7 @@ use pango::prelude::FontMapExt;
 use regex::{Captures, Regex};
 use std::borrow::Cow;
 use std::cell::RefCell;
+use std::convert::TryFrom;
 use std::f64::consts::*;
 use std::rc::Rc;
 
@@ -182,6 +183,8 @@ pub struct DrawingCtx {
 
     drawsub_stack: Vec<Node>,
 
+    svg_nesting: SvgNesting,
+
     measuring: bool,
     testing: bool,
 }
@@ -192,6 +195,20 @@ pub enum DrawingMode {
     OnlyNode(Node),
 }
 
+/// Whether an SVG document is being rendered standalone or referenced from an `<image>` element.
+///
+/// Normally, the coordinate system used when rendering a toplevel SVG is determined from the
+/// initial viewport and the `<svg>` element's `viewBox` and `preserveAspectRatio` attributes.
+/// However, when an SVG document is referenced from an `<image>` element, as in `<image href="foo.svg"/>`,
+/// its `preserveAspectRatio` needs to be ignored so that the one from the `<image>` element can
+/// be used instead.  This lets the parent document (the one with the `<image>` element) specify
+/// how it wants the child SVG to be scaled into the viewport.
+#[derive(Copy, Clone)]
+pub enum SvgNesting {
+    Standalone,
+    ReferencedFromImageElement,
+}
+
 /// The toplevel drawing routine.
 ///
 /// This creates a DrawingCtx internally and starts drawing at the specified `node`.
@@ -202,6 +219,7 @@ pub fn draw_tree(
     viewport_rect: Rect,
     user_language: &UserLanguage,
     dpi: Dpi,
+    svg_nesting: SvgNesting,
     measuring: bool,
     testing: bool,
     acquired_nodes: &mut AcquiredNodes<'_>,
@@ -255,6 +273,7 @@ pub fn draw_tree(
         &initial_viewport,
         user_language.clone(),
         dpi,
+        svg_nesting,
         measuring,
         testing,
         drawsub_stack,
@@ -303,6 +322,7 @@ impl DrawingCtx {
         initial_viewport: &Viewport,
         user_language: UserLanguage,
         dpi: Dpi,
+        svg_nesting: SvgNesting,
         measuring: bool,
         testing: bool,
         drawsub_stack: Vec<Node>,
@@ -315,6 +335,7 @@ impl DrawingCtx {
             cr: cr.clone(),
             user_language,
             drawsub_stack,
+            svg_nesting,
             measuring,
             testing,
         }
@@ -339,6 +360,7 @@ impl DrawingCtx {
             cr,
             user_language: self.user_language.clone(),
             drawsub_stack: self.drawsub_stack.clone(),
+            svg_nesting: self.svg_nesting,
             measuring: self.measuring,
             testing: self.testing,
         }
@@ -379,10 +401,18 @@ impl DrawingCtx {
         }
     }
 
+    pub fn svg_nesting(&self) -> SvgNesting {
+        self.svg_nesting
+    }
+
     pub fn is_measuring(&self) -> bool {
         self.measuring
     }
 
+    pub fn is_testing(&self) -> bool {
+        self.testing
+    }
+
     pub fn get_transform(&self) -> ValidTransform {
         let t = Transform::from(self.cr.matrix());
         ValidTransform::try_from(t)
@@ -2064,7 +2094,11 @@ fn compute_stroke_and_fill_extents(
     // rectangle's extents, even when it has no fill nor stroke.
 
     let (x0, y0, x1, y1) = cr.fill_extents()?;
-    let fill_extents = Some(Rect::new(x0, y0, x1, y1));
+    let fill_extents = if x0 != 0.0 || y0 != 0.0 || x1 != 0.0 || y1 != 0.0 {
+        Some(Rect::new(x0, y0, x1, y1))
+    } else {
+        None
+    };
 
     // Bounding box for stroke
     //
