@@ -20,7 +20,7 @@
 //!
 //! ```toml
 //! [dependencies]
-//! librsvg = "2.57.4"
+//! librsvg = "2.58.0-beta.0"
 //! cairo-rs = "0.18.2"
 //! gio = "0.18"   # only if you need streams
 //! ```
@@ -166,7 +166,6 @@ mod filters;
 mod float_eq_cairo;
 mod font_props;
 mod gradient;
-mod handle;
 mod href;
 mod image;
 mod io;
@@ -225,7 +224,6 @@ pub mod bench_only {
 #[cfg(feature = "c-api")]
 pub mod c_api_only {
     pub use crate::dpi::Dpi;
-    pub use crate::handle::Handle;
     pub use crate::rsvg_log;
     pub use crate::session::Session;
     pub use crate::surface_utils::shared_surface::{SharedImageSurface, SurfaceType};
