@@ -4,6 +4,7 @@ use std::marker::PhantomData;
 use std::ptr::NonNull;
 use std::slice;
 
+use cast::i32;
 use cssparser::Color;
 use gdk_pixbuf::{Colorspace, Pixbuf};
 use nalgebra::{storage::Storage, Dim, Matrix};
@@ -242,9 +243,7 @@ impl ImageSurface<Shared> {
         // Cairo allows zero-sized surfaces, but it does malloc(0), whose result
         // is implementation-defined.  So, we can't assume NonNull below.  This is
         // why we disallow zero-sized surfaces here.
-        if !(width > 0 && height > 0) {
-            return Err(cairo::Error::InvalidSize);
-        }
+        assert!(width > 0 && height > 0);
 
         surface.flush();
 
@@ -304,53 +303,6 @@ impl ImageSurface<Shared> {
         }
     }
 
-    pub fn from_pixbuf(
-        pixbuf: &Pixbuf,
-        content_type: Option<&str>,
-        mime_data: Option<Vec<u8>>,
-    ) -> Result<SharedImageSurface, cairo::Error> {
-        assert!(pixbuf.colorspace() == Colorspace::Rgb);
-        assert!(pixbuf.bits_per_sample() == 8);
-
-        let n_channels = pixbuf.n_channels();
-        assert!(n_channels == 3 || n_channels == 4);
-        let has_alpha = n_channels == 4;
-
-        let width = pixbuf.width();
-        let height = pixbuf.height();
-        let stride = pixbuf.rowstride() as usize;
-        assert!(width > 0 && height > 0 && stride > 0);
-
-        let pixbuf_data = unsafe { pixbuf.pixels() };
-
-        let mut surf = ExclusiveImageSurface::new(width, height, SurfaceType::SRgb)?;
-
-        // We use chunks(), not chunks_exact(), because gdk-pixbuf tends
-        // to make the last row *not* have the full stride (i.e. it is
-        // only as wide as the pixels in that row).
-        let pixbuf_rows = pixbuf_data.chunks(stride).take(height as usize);
-
-        if has_alpha {
-            pixbuf_rows
-                .map(|row| row.as_rgba())
-                .zip(surf.rows_mut())
-                .flat_map(|(src_row, dest_row)| src_row.iter().zip(dest_row.iter_mut()))
-                .for_each(|(src, dest)| *dest = src.to_pixel().premultiply().to_cairo_argb());
-        } else {
-            pixbuf_rows
-                .map(|row| row.as_rgb())
-                .zip(surf.rows_mut())
-                .flat_map(|(src_row, dest_row)| src_row.iter().zip(dest_row.iter_mut()))
-                .for_each(|(src, dest)| *dest = src.to_pixel().to_cairo_argb());
-        }
-
-        if let (Some(content_type), Some(bytes)) = (content_type, mime_data) {
-            surf.surface.set_mime_data(content_type, bytes)?;
-        }
-
-        surf.share()
-    }
-
     pub fn to_pixbuf(&self) -> Option<Pixbuf> {
         let width = self.width();
         let height = self.height();
@@ -378,6 +330,31 @@ impl ImageSurface<Shared> {
         Some(pixbuf)
     }
 
+    pub fn from_image(
+        image: &image::DynamicImage,
+        content_type: Option<&str>,
+        mime_data: Option<Vec<u8>>,
+    ) -> Result<SharedImageSurface, cairo::Error> {
+        let rgba_image = image.to_rgba8();
+
+        let width = i32(rgba_image.width()).map_err(|_| cairo::Error::InvalidSize)?;
+        let height = i32(rgba_image.height()).map_err(|_| cairo::Error::InvalidSize)?;
+
+        let mut surf = ExclusiveImageSurface::new(width, height, SurfaceType::SRgb)?;
+
+        rgba_image
+            .rows()
+            .zip(surf.rows_mut())
+            .flat_map(|(src_row, dest_row)| src_row.zip(dest_row.iter_mut()))
+            .for_each(|(src, dest)| *dest = src.to_pixel().premultiply().to_cairo_argb());
+
+        if let (Some(content_type), Some(bytes)) = (content_type, mime_data) {
+            surf.surface.set_mime_data(content_type, bytes)?;
+        }
+
+        surf.share()
+    }
+
     /// Returns `true` if the surface contains meaningful data only in the alpha channel.
     #[inline]
     fn is_alpha_only(&self) -> bool {
@@ -1037,7 +1014,7 @@ impl ImageSurface<Shared> {
     #[inline]
     pub fn offset(
         &self,
-        bounds: Rect,
+        bounds: IRect,
         dx: f64,
         dy: f64,
     ) -> Result<SharedImageSurface, cairo::Error> {
@@ -1046,7 +1023,10 @@ impl ImageSurface<Shared> {
 
         // output_bounds contains all pixels within bounds,
         // for which (x - ox) and (y - oy) also lie within bounds.
-        if let Some(output_bounds) = bounds.translate((dx, dy)).intersection(&bounds) {
+        if let Some(output_bounds) = bounds
+            .translate((dx as i32, dy as i32))
+            .intersection(&bounds)
+        {
             let cr = cairo::Context::new(&output_surface)?;
             let r = cairo::Rectangle::from(output_bounds);
             cr.rectangle(r.x(), r.y(), r.width(), r.height());
@@ -1356,9 +1336,7 @@ impl ImageSurface<Exclusive> {
         // Cairo allows zero-sized surfaces, but it does malloc(0), whose result
         // is implementation-defined.  So, we can't assume NonNull below.  This is
         // why we disallow zero-sized surfaces here.
-        if !(width > 0 && height > 0) {
-            return Err(cairo::Error::InvalidSize);
-        }
+        assert!(width > 0 && height > 0);
 
         let data_ptr = NonNull::new(unsafe {
             cairo::ffi::cairo_image_surface_get_data(surface.to_raw_none())
@@ -1458,6 +1436,7 @@ impl From<Operator> for cairo::Operator {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::surface_utils::iterators::Pixels;
 
     #[test]
     fn test_extract_alpha() {
