@@ -21,9 +21,8 @@ use xml5ever::tokenizer::{TagKind, Token, TokenSink, XmlTokenizer, XmlTokenizerO
 
 use crate::borrow_element_as;
 use crate::css::{Origin, Stylesheet};
-use crate::document::{Document, DocumentBuilder};
+use crate::document::{Document, DocumentBuilder, LoadOptions};
 use crate::error::{ImplementationLimit, LoadingError};
-use crate::handle::LoadOptions;
 use crate::io::{self, IoError};
 use crate::limits::{MAX_LOADED_ELEMENTS, MAX_XINCLUDE_DEPTH};
 use crate::node::{Node, NodeBorrow};
