@@ -634,7 +634,7 @@ impl Converter {
                         }
                     };
 
-                    // Supported SVG units are px, in, cm, mm, pt, pc
+                    // Supported SVG units are px, in, cm, mm, pt, pc, ch
                     (
                         Size {
                             w: set_unit(
@@ -1214,6 +1214,33 @@ fn parse_zoom_factor(v: &str) -> Result<ZoomFactor, String> {
     }
 }
 
+trait NotFound {
+    type Ok;
+    type Error;
+
+    fn or_none(self) -> Result<Option<Self::Ok>, Self::Error>;
+}
+
+impl<T> NotFound for Result<T, clap::Error> {
+    type Ok = T;
+    type Error = clap::Error;
+
+    /// Maps the Result to an Option, translating the ArgumentNotFound error to
+    /// Ok(None), while mapping other kinds of errors to Err(e).
+    ///
+    /// This allows to get proper error reporting for invalid values on optional
+    /// arguments.
+    fn or_none(self) -> Result<Option<T>, clap::Error> {
+        self.map_or_else(
+            |e| match e.kind() {
+                clap::error::ErrorKind::UnknownArgument => Ok(None),
+                _ => Err(e),
+            },
+            |v| Ok(Some(v)),
+        )
+    }
+}
+
 fn parse_background_color(s: &str) -> Result<Option<Color>, String> {
     match s {
         "none" | "None" => Ok(None),
@@ -1227,7 +1254,7 @@ fn is_absolute_unit(u: LengthUnit) -> bool {
     use LengthUnit::*;
 
     match u {
-        Percent | Em | Ex => false,
+        Percent | Em | Ex | Ch => false,
         Px | In | Cm | Mm | Pt | Pc => true,
     }
 }
